from collections.abc import Generator


class Input:
    """
    Input can be either a file or any other type
    """

    def __init__(self, value: str = "", file: bool = False) -> None:
        self.value = value
        self.file = file

    def __iter__(self) -> Generator[tuple, tuple, None]:
        yield "value", self.value
        yield "file", self.file



import os
import time
import sys
import warnings
from importlib.metadata import version
from pathlib import Path
from typing import Callable
from frelatage.corpus.corpus import load_corpus
from frelatage.config.config import Config
from frelatage.input.input import Input
from frelatage.mutator.mutator import *
from frelatage.queue.queue import Queue
from frelatage.tracer.tracer import Tracer

# Automatically retrieves the current version number
__version__ = version("frelatage")
__author__ = "Rog3rSm1th"
__email__ = "r0g3r5@protonmail.com"
__git__ = "https://github.com/Rog3rSm1th"
__twitter__ = "https://twitter.com/Rog3rSm1th"


class Fuzzer(object):
    """
    Copyright (c) 2022 Rog3rSm1th

    Frelatage is a coverage-based Python fuzzing library which can be used to fuzz python code.
    The development of Frelatage was inspired by various other fuzzers, including AFL/AFL++,
    Atheris and PyFuzzer.The main purpose of the project is to take advantage of the best features
    of these fuzzers and gather them together into a new tool in order to efficiently fuzz python applications.
    """

    # List of all defined harnesses
    instances: list = []

    from ._mutation import valid_mutators, get_mutation, generate_cycle_mutations  # type: ignore
    from ._interface import (  # type: ignore
        init_interface,
        refresh_interface,
        start_interface,
        exit_message,
        kill_interface,
    )
    from ._evaluate import evaluate_mutations  # type: ignore
    from ._cycle import run_function, run_cycle  # type: ignore
    from ._fuzz import fuzz  # type: ignore
    from ._report import get_report_name, save_report  # type: ignore
    from ._input import init_input_folder, init_file_input_arguments, init_file_inputs  # type: ignore

    def __init__(
        self,
        method: Callable,
        corpus: list,
        threads_count: int = 8,
        exceptions_whitelist: tuple = (),
        exceptions_blacklist: tuple = (),
        output_directory: str = "./out",
        coverage_directory: str = "./cov",
        silent: bool = False,
        infinite_fuzz: bool = False,
        report: bool = True,
    ) -> None:
        """
        Initialize the fuzzer
        """
        warnings.filterwarnings("ignore")

        self.__class__.instances.append(self)

        # Status of the fuzzer activity
        self.alive = True

        # Frelatage configuration
        self.config = Config

        # Global set of reached instructions
        self.reached_instructions: set = set([])
        # Global set of instructions pairs
        self.instructions_pairs: set = set([])
        # Global set of instruction pairs executed during a crash
        self.favored_pairs: set = set([])
        # Global Set of positions where a crash occurred
        self.error_positions: set = set([])
        # Fuzzed method
        self.method = method

        # Remove duplicates in corpus
        corpus = [list(set(argument)) for argument in corpus]
        # Frelatage corpus
        self.corpus = corpus
        # List of all avalaibles mutators
        self.mutators = mutators
        # Number of concurrently launched threads
        self.threads_count = max(min(threads_count, Config.FRELATAGE_MAX_THREADS), 8)
        # List of cycle mutations
        self.cycle: list = []

        # Exceptions that will be taken into account or not when fuzzing
        self.exceptions_whitelist = exceptions_whitelist
        self.exceptions_blacklist = exceptions_blacklist

        # Frelatage tracer
        self.tracer = Tracer(
            exceptions_whitelist=self.exceptions_whitelist,
            exceptions_blacklist=self.exceptions_blacklist,
        )

        # Input, output and coverage directories
        # The working directory is the same as the fuzz file
        self.input_directory = os.path.join(
            os.path.dirname(os.path.realpath(sys.argv[0])),
            Config.FRELATAGE_INPUT_DIR,
        )
        self.output_directory = Path(
            os.path.join(
                os.path.dirname(os.path.realpath(sys.argv[0])),
                output_directory,
            )
        ).as_posix()
        self.coverage_directory = Path(
            os.path.join(
                os.path.dirname(os.path.realpath(sys.argv[0])),
                coverage_directory,
            )
        ).as_posix()

        # Silent output
        self.silent = silent

        # Infinite fuzzing
        self.infinite_fuzz = infinite_fuzz

        # Fuzzer statistics
        self.cycles_count = 0
        self.inputs_count = 0
        self.stage_inputs_count = 0
        self.unique_crashes = 0
        self.unique_coverage_increase = 0
        self.total_crashes = 0
        self.unique_timeout = 0
        self.total_timeouts = 0

        # Time statistics
        self.fuzz_start_time = None
        self.last_new_path_time = None
        self.last_unique_crash_time = None
        self.last_unique_timeout_time = None

        # Genetic algorithm parameters
        self.survival_probability = 0.8
        self.mutation_probability = 0.2

        # Number of Frelatage cycles without finding new paths
        self.cycles_without_new_path = 0
        # Corpus entries that are still in the queue
        self.queue = Queue(self.corpus)
        # Check if the queue has a reasonable size
        if self.queue.size > Config.FRELATAGE_MAX_STAGES:
            raise ValueError(
                "The number of stages is too high for method {method} ({stages_count}), max is ({config_max_stages})".format(
                    method=self.method.__name__,
                    stages_count=str(self.queue.size),
                    config_max_stages=str(Config.FRELATAGE_MAX_STAGES),
                )
            )

        # Current arguments
        self.arguments = self.queue.current_arguments()

        # Executions per second
        self.last_seconds_executions = 0
        self.executions_per_second = 0
        self.current_second = int(time.time())

        # Print or not a report at the end of the fuzzing
        self.report = report

    @staticmethod
    def fuzz_all() -> None:
        """
        Fuzzing of all instrumented functions
        """
        for instance in Fuzzer.instances:
            instance.fuzz()


def instrument(
    corpus: list,
    threads_count: int = 8,
    exceptions_whitelist: tuple = (),
    exceptions_blacklist: tuple = (),
    output_directory: str = "./out",
    silent: bool = False,
    infinite_fuzz: bool = False,
) -> Callable:
    """
    Set up a fuzzer around a function
    """
    decorator_arguments = list(locals().values())

    def _decorate(function: Any) -> None:
        report = False
        harness_arguments: List[Any] = [function] + decorator_arguments + [report]
        f = Fuzzer(*harness_arguments)

    return _decorate



import copy
import os
import random
from pathlib import Path
from typing import Type, Any
from frelatage.mutator.mutator import Mutator
from pympler.asizeof import asizeof


def get_argument_size(argument, file: bool) -> int:
    """
    Return the argument size in bytes
    """
    if file:
        # File size
        size = os.path.getsize(argument)
    else:
        # Variable size in memory
        size = asizeof(argument)
    return size


def valid_mutators(self, input_type: Type, input_size: int) -> list[Mutator]:
    """
    Filters mutators by input type
    """
    # Argument type
    # e.g. "int", "str", "file"
    type_str = str(input_type.__name__) if input_type != "file" else "file"
    # Filtering mutators using the "allowed types" property of each mutator
    valid_mutators_type = list(
        filter(
            lambda mutator: type_str in mutator.allowed_types and mutator.enabled,
            self.mutators,
        )
    )

    # Prevents the input from increasing in size if the maximum allowed size has been reached
    max_input_size_reached = input_size > self.config.FRELATAGE_INPUT_MAX_LEN
    if max_input_size_reached:
        # Filtering mutators using the "size_effect" property of each mutator
        valid_mutators_size = list(
            filter(
                lambda mutator: "decrease" in mutator.size_effect,
                self.mutators,
            )
        )
    else:
        valid_mutators_size = self.mutators

    # Filtered mutators
    valid_mutators = list(set(valid_mutators_type) & set(valid_mutators_size))
    return valid_mutators


def get_mutation(self, input: Any, file: bool) -> Any:
    """
    Generate a mutation in a recursive way for a variable
    e.g. : get_mutation(["a", 1, ["b"]])
            -> We start by mutating the global array
            -> Then we recursively mutate the elements of the array until
            all non-iterable elements have been reached. (In this particular
            case : "a", 1 and "b")
    """
    mutate = random.randint(0, 100) / 100 <= self.mutation_probability

    # If we mutate the input
    if mutate:
        input_type = type(input) if not file else "file"
        is_file = True if input_type == "file" else False

        argument_size = get_argument_size(input, is_file)
        valid_mutators = self.valid_mutators(input_type, argument_size)
        # # We use a random mutator among the valid ones
        mutator = random.choice(valid_mutators)
        # Root element mutation
        mutation = mutator.mutate(copy.deepcopy(input))

        is_list = isinstance(mutation, list)
        is_tuple = isinstance(mutation, tuple)
        is_iterable = is_list or is_tuple
        is_dict = isinstance(mutation, dict)

        # If the element is iterable we mutate all its children elements
        if is_iterable:
            # Convert the tuple into a list so that we can modify the elements
            mutation = list(mutation) if is_tuple else mutation
            for i in range(len(mutation)):
                valid_mutators = self.valid_mutators(type(mutation[i]), 1)
                mutation[i] = self.get_mutation(mutation[i], False)
            # Reconvert to a tuple
            mutation = tuple(mutation) if is_tuple else mutation

        elif is_dict:
            # Mutate keys
            for key in list(mutation.keys()):
                new_key = self.get_mutation(key, False)
                mutation[new_key] = mutation.pop(key)

            # Mutate values
            for key in mutation:
                valid_mutators = self.valid_mutators(type(mutation[key]), 1)
                mutation[key] = self.get_mutation(mutation[key], False)

    # If we don't mutate the input
    else:
        mutation = input
    return mutation


def generate_cycle_mutations(self, parents: list) -> None:
    """
    Generate a list of mutations with a list of parent mutations as an input
    """
    new_cycle = []

    # Mutation
    for thread in range(self.threads_count):

        # Selecting a random parent
        arguments = random.choice(parents)

        thread_arguments = []

        # We mutate each argument
        for argument in arguments:

            mutation = copy.deepcopy(argument)
            mutation.value = self.get_mutation(mutation.value, mutation.file)
            # Mutation of "file" type inputs
            if mutation.file:
                filename = os.path.split(mutation.value)[1]
                file_argument_id = os.path.split(os.path.split(mutation.value)[0])[1]
                base = Path(mutation.value).parents[2]
                new_argument = os.path.join(base, str(thread), file_argument_id, filename)
                mutation.value = new_argument

            thread_arguments.append(mutation)
        new_cycle.append(thread_arguments)
    self.cycle = new_cycle



from datetime import datetime
from threading import Thread
from frelatage.config.config import Config


def fuzz(self) -> None:
    """
    Run the fuzzer
    """
    # Initialize the fuzzer timer
    self.fuzz_start_time = datetime.now()

    # Initialize file input folders in /tmp/frelatage (default value)
    # Can be modified using the FRELATAGE_INPUT_FILE_TMP_DIR env variable
    self.init_file_inputs()

    # Infinite fuzzing is allowed if we have one input combination
    if self.infinite_fuzz and len(self.queue.arguments) > 1:
        print("Error: infinite fuzzing is only possible with a corpus of size 1")
        exit(1)

    try:
        # Interface
        if not self.silent:
            p = Thread(target=self.start_interface)
            p.daemon = True
            p.start()

        # Fuzzing
        parents = [self.arguments]
        while True:
            self.generate_cycle_mutations(parents)
            reports = self.run_cycle()
            # If no new paths have been found for a while, we go to the next stage
            if (
                self.cycles_without_new_path
                >= Config.FRELATAGE_MAX_CYCLES_WITHOUT_NEW_PATHS
                and not self.infinite_fuzz
            ):
                # Next stage
                self.queue.position += 1
                if not self.queue.end:
                    # Initialize the new stage
                    self.arguments = self.queue.current_arguments()
                    parents = [self.arguments]
                    self.cycles_without_new_path = 0
                    self.stage_inputs_count = 0
                    self.init_file_inputs()
                # End of the fuzzing process
                # Exit the program
                else:
                    if self.report:
                        self.exit_message(normal_ending=True)
                        exit(1)
                    self.alive = False
                    self.kill_interface()
                    break
            else:
                parents = self.evaluate_mutations(reports)
    # Exit the program
    # Keyboard interrupt
    except KeyboardInterrupt:
        self.exit_message(aborted_by_user=True)
        exit(0)
    # Error in Frelatage
    except Exception as e:
        self.exit_message(aborted_by_user=False)
        if Config.FRELATAGE_DEBUG_MODE:
            print(e)
        exit(1)



import curses
import time
from curses import wrapper
from datetime import datetime, timedelta
from string import Formatter
from frelatage import __version__, Config
from frelatage.colors import Colors

# Refresh the interface 10 times/second
REFRESH_INTERVAL = 0.1


def format_delta(time_delta: timedelta, format: str) -> str:
    """
    Format a time delta.
    """
    formatter = Formatter()
    digits = {}
    constants = {"D": 86400, "H": 3600, "M": 60, "S": 1}
    k = map(lambda x: x[1], list(formatter.parse(format)))
    remaining = int(time_delta.total_seconds())

    for i in ("D", "H", "M", "S"):
        if i in k and i in constants.keys():
            digits[i], remaining = divmod(remaining, constants[i])

    return formatter.format(format, **digits)


def format_time_elapsed(datetime: datetime) -> str:
    """
    Format the time elapsed since a datetime in the format :
    {D} days, {H} hrs, {M} min, {S} sec
    """
    if datetime is not None:
        format_time_delta = datetime.now() - datetime
        format_time = format_delta(
            format_time_delta, "{D} days, {H} hrs, {M} min, {S} sec"
        )
    # If the event has not yet occurred
    else:
        format_time = "none seen yet"
    return format_time


def init_interface(self, stdscr) -> bool:
    """
    Initialize the Frelatage CLI
    """
    # Hide cursor
    curses.curs_set(0)
    # Colors
    curses.start_color()
    curses.use_default_colors()

    # Adapt the colors depending on the terminal type
    try:
        # xterm-256color
        for i in range(0, 255):
            curses.init_pair(i + 1, i, -1)
    except:
        # xterm-color
        curses.init_pair(1, curses.COLOR_WHITE, curses.COLOR_BLACK)

    self.screen = stdscr
    self.screen.clear()
    self.screen.refresh()
    return True


def refresh_interface(self):
    """
    Refresh the Frelatage CLI
    """
    # Title
    title = "Frelatage {version} ({function_name})".format(
        version=__version__, function_name=self.method.__name__
    ).center(105)

    # Execs per second
    current_second = int(time.time())
    if current_second != self.current_second:
        self.executions_per_second = self.last_seconds_executions
        self.last_seconds_executions = 0
        self.current_second = int(time.time())
    execs_per_second = str(self.executions_per_second).rjust(6)

    # Process timing
    run_time = format_time_elapsed(self.fuzz_start_time).ljust(32)
    last_new_path_time = format_time_elapsed(self.last_new_path_time).ljust(32)
    last_unique_crash_time = format_time_elapsed(self.last_unique_crash_time).ljust(32)
    last_unique_timeout_time = format_time_elapsed(self.last_unique_timeout_time).ljust(
        32
    )

    # Overall results
    uniques_crashes_count = str(self.unique_crashes).ljust(9)
    uniques_timeouts_count = str(self.unique_timeout).ljust(9)

    # Finding in depth
    total_paths_count = str(len(self.reached_instructions)).ljust(22)
    favored_paths_count = len(self.favored_pairs)
    favored_paths_rate = (
        round(int(favored_paths_count) / int(total_paths_count) * 100, 2)
        if int(total_paths_count)
        else 0.00
    )
    favored_paths = "{favored_paths} ({rate}%)".format(
        favored_paths=favored_paths_count, rate=favored_paths_rate
    ).ljust(22)

    # Crashes
    total_crashes = "{crashes} ({uniques} uniques)".format(
        crashes=str(self.total_crashes), uniques=str(self.unique_crashes)
    ).ljust(22)
    total_timeouts = "{total_timeouts} [{timeout_delay} sec]".format(
        total_timeouts=self.total_timeouts,
        timeout_delay=Config.FRELATAGE_TIMEOUT_DELAY,
    ).ljust(22)

    # Progress
    cycles_count = str(self.cycles_count).ljust(12)
    total_executions = str(self.inputs_count).ljust(12)

    # Stage progress
    current_argument = self.queue.position + 1
    total_arguments_count = len(self.queue.arguments)
    current_stage = "{current_argument}/{total_arguments_count}".format(
        current_argument=current_argument,
        total_arguments_count=total_arguments_count,
    ).ljust(16)
    stage_executions = str(self.stage_inputs_count).ljust(16)

    # Interface
    self.screen.addstr(
        0,
        0,
        """
    {title}

    ┌──── Process timing ────────────────────────────────────┬─────── Finding in depth ─────────────────────┐
    │ Run time            :: {run_time}│ Favored paths       :: {favored_paths}│
    │ Last new path       :: {last_new_path_time}│ Total paths         :: {total_paths_count}│
    │ Last unique crash   :: {last_unique_crash_time}│ Total timeouts      :: {total_timeouts}│
    │ Last unique timeout :: {last_unique_timeout_time}│ Total crashes       :: {total_crashes}│
    ├──── Overall result ─────────────┬─── Global progress ──┴─────────────┬──── Stage progress─────────────┤
    │ Uniques crashes     :: {uniques_crashes_count}│ Cycles done         :: {cycles_count}│ Stage       :: {current_stage}│
    │ Unique timeouts     :: {uniques_timeouts_count}│ Total executions    :: {total_executions}│ Stage execs :: {stage_executions}│
    └─────────────────────────────────┴────────────────────────────────────┴────────────────────────────────┘
    [ {execs_per_second} exec/s ]                             
    """.format(
            title=title,
            execs_per_second=execs_per_second,
            run_time=run_time,
            last_new_path_time=last_new_path_time,
            last_unique_crash_time=last_unique_crash_time,
            last_unique_timeout_time=last_unique_timeout_time,
            uniques_crashes_count=uniques_crashes_count,
            uniques_timeouts_count=uniques_timeouts_count,
            favored_paths=favored_paths,
            total_paths_count=total_paths_count,
            total_timeouts=total_timeouts,
            total_crashes=total_crashes,
            cycles_count=cycles_count,
            total_executions=total_executions,
            current_stage=current_stage,
            stage_executions=stage_executions,
        ),
    )
    self.screen.refresh()


def exit_message(
    self, normal_ending: bool = False, aborted_by_user: bool = False
) -> bool:
    """
    Message displayed when exiting the program
    """
    run_time = format_time_elapsed(self.fuzz_start_time)
    uniques_crashes_count = str(self.unique_crashes)
    uniques_timeouts_count = str(self.unique_timeout)
    total_timeouts = "{total_timeouts}".format(total_timeouts=self.total_timeouts)
    total_paths_count = str(len(self.reached_instructions))
    cycles_count = str(self.cycles_count)
    total_executions = str(self.inputs_count)

    # End the curse window
    if not self.silent:
        curses.endwin()

    # Keyboard interrupt
    if aborted_by_user:
        print(Colors.FAIL + "+++ Fuzzing aborted by user +++" + "\r\n")
    # Normal ending
    elif normal_ending:
        print(Colors.OKGREEN + "+++ Fuzzing completed +++" + "\r\n")
    # Error in the program
    else:
        print(
            Colors.FAIL
            + "+++ Fuzzing was interrupted by an error in Frelatage +++"
            + "\r\n"
        )

    # Message displayed at the end of the program
    print(
        Colors.OKGREEN
        + "[+] "
        + Colors.ENDC
        + "Error reports are located in: "
        + Colors.BOLD
        + self.output_directory
        + Colors.ENDC
        + "\r\n"
        + "\r\n"
        + Colors.BOLD
        + "Fuzzing statistics: "
        + Colors.ENDC
        + "\r\n"
        + "\r\n"
        + "Total run time: "
        + run_time
        + "\r\n"
        + "Uniques crashes: "
        + Colors.FAIL
        + uniques_crashes_count
        + Colors.ENDC
        + "\r\n"
        + "Uniques timeouts: "
        + uniques_timeouts_count
        + "\r\n"
        + "Crashes: "
        + str(self.total_crashes)
        + "\r\n"
        + "Timeouts: "
        + total_timeouts
        + "\r\n"
        + "Reached paths: "
        + total_paths_count
        + "\r\n"
        + "Cycles: "
        + cycles_count
        + "\r\n"
        + "Executions: "
        + total_executions
        + "\r"
    )
    return True


def start_interface(self) -> None:
    """
    Display the curse interface
    """
    wrapper(self.init_interface)
    while self.alive:
        self.refresh_interface()
        time.sleep(REFRESH_INTERVAL)


def kill_interface(self) -> None:
    """
    Kill the current curses window
    """
    curses.endwin()



from frelatage.report.report import Report


def run_function(self, arguments: list, result: list) -> bool:
    """
    Run a function inside the tracer and return a report
    """
    trace = self.tracer.trace(self.method, arguments)

    trace_instructions = set([instruction for instruction in trace.reached_instructions])
    trace_instruction_pairs = set([pair for pair in trace.instructions_pairs])

    trace_instructions_count = len(trace_instructions)
    trace_instruction_pairs_count = len(trace_instruction_pairs)

    new_instructions_count = len(trace_instructions - self.reached_instructions)
    new_sequences_count = len(trace_instruction_pairs - self.instructions_pairs)

    new_instruction_error = trace.error_position not in self.error_positions

    run_report = Report(
        trace,
        arguments,
        trace.error,
        trace.timeout,
        new_instruction_error,
        trace_instructions_count,
        new_instructions_count,
        trace_instruction_pairs_count,
        new_sequences_count,
    )
    result.append(run_report)
    return True


def run_cycle(self) -> list[Report]:
    """
    run the function with each mutation of the cycle as argument.
    Return the execution reports.
    """
    cycle_reports: list = []

    # TODO: Implement multithreading
    for mutation in self.cycle:
        self.run_function(mutation, cycle_reports)
        self.inputs_count += 1
        self.last_seconds_executions += 1
        self.stage_inputs_count += 1
    self.cycles_count += 1
    return cycle_reports



#!/usr/bin/env python3
# usage : ./read_report input
import pickle
import sys


def main():
    if len(sys.argv) > 1:
        report_file = sys.argv[1]
    else:
        print("Usage : ./read_report.py <report file>")
        sys.exit(1)

    with open(report_file, "rb") as f:
        report = pickle.load(f)
        print(report)



import os
import pickle
from frelatage.report.report import Report


def get_report_name(self, report: Report) -> str:
    """
    Generate a report title.
    The name of a report is in the following form:
    id:<crash ID>,err:<error type>,err_pos:<error>,err_file:<error file>,err_po:<err_pos>
    """
    # It is assumed that the number of uniques crashes will not exceed 999999
    error_id = str(self.unique_crashes).zfill(6)
    # Type of the error in lowercase
    error_type = report.trace.error_type.lower()
    # Line of the file where the error occured
    error_position = (
        str(report.trace.error_position[0][1])
        if report.trace.error_position is not None
        else str(None)
    )
    # File where the error occured
    error_file = os.path.splitext(os.path.basename(report.trace.error_position[0][0]))[
        0
    ].lower()

    report_name = "id:{error_id},err:{error_type},err_file:{error_file},err_pos:{error_position}".format(
        error_id=error_id,
        error_type=error_type,
        error_file=error_file,
        error_position=error_position,
    )
    return report_name


def save_report(self, report: Report, crash: bool = True) -> bool:
    """
    Save a report in the output directory (/out by default).
    The arguments passed to the function and the behavior of the function are stored
    in "input", the file inputs are stored in files ranging from 0 to n.
    The report directory is in the following form:
    ├── out
    │   ├── id:<crash ID>,err:<error type>,err_pos:<error>,err_file:<error file>,err_pos:<err_pos>
    │       ├── input
    │       ├── 0
    │           ├── <inputfile1>
    │       ├── ...
    │   ├── ...
    """

    # The report contains the parameters passed to the function.
    custom_report = {"input": [dict(input) for input in report.input]}

    # ./out by default if it"s a crash report
    # ./cov by default if it's a coverage increase report
    base_directory = self.output_directory if crash else self.coverage_directory
    report_name = (
        self.get_report_name(report)
        if crash
        else "coverage_{}".format(self.unique_coverage_increase)
    )

    report_directory = "{base_directory}/{report_name}".format(
        base_directory=base_directory, report_name=report_name
    )

    # Create ./out/<report directory>/ directory if not exists
    if not os.path.exists(report_directory):
        os.makedirs(report_directory)

    # Save report file
    with open(
        "{report_directory}/input".format(report_directory=report_directory),
        "wb+",
    ) as f:
        # We use pickle to store the report object
        pickle.dump(custom_report, f)

    # Save input files
    for i in range(len(report.input)):
        argument = report.input[i]
        argument_number = i

        if argument.file:
            file_argument_content = open(argument.value, "rb").read()
            # Save file in /out/<report name>/<argument number>/<file name>
            argument_directory = "{report_directory}/{argument_number}".format(
                report_directory=report_directory,
                argument_number=argument_number,
            )
            # create /out/<report name>/<argument number> folder if not exists
            if not os.path.exists(argument_directory):
                os.makedirs(argument_directory)

            filename = os.path.basename(argument.value)
            with open(
                "{argument_directory}/{filename}".format(
                    argument_directory=argument_directory, filename=filename
                ),
                "wb+",
            ) as f:
                f.write(file_argument_content)
    return True



import random
from typing import List, Any, Type
from frelatage.config.config import Config
from frelatage.mutator.dictionary import load_dictionary
from frelatage.mutator.magicValues import MagicValues

# Array containing all the mutators
mutators: List[Any] = []

# Load dictionary from the dictionary files
dictionary_folder = Config().FRELATAGE_DICTIONARY_DIR
dictionary = load_dictionary(dictionary_folder)


class Mutator(object):
    # "str", "int", "list", "dict", "NoneType", "float"
    allowed_types: set = set([])
    # "none", "increase", "decrease"
    size_effect: list = []
    # Is the mutator used by the fuzzer
    # True by default
    enabled = True

    @staticmethod
    def random_int(max: int) -> int:
        """
        Return a random integer.
        """
        if max == 1 or max == 0:
            return 0
        return random.randint(0, max - 1)

    @staticmethod
    def random_magic_value(magic_values_set: list) -> Any:
        """
        Return a random value from a magic value array.
        """
        return random.choice(magic_values_set)

    def mutate(self, input: Any) -> Any:
        """
        Function to mutate a given resource into another one.
        @return: new resource, or None if this mutator is not appropriate.
        """
        raise NotImplementedError(
            "mutate not implemented in {}".format(self.__class__.__name__)
        )


def register_mutator(mutator: Type[Mutator]) -> bool:
    """
    Register a mutator into the global mutators array
    """
    mutators.append(mutator)
    return True


# String mutators
@register_mutator
class MutatorStringBitFlip(Mutator):
    allowed_types = set(["str"])
    size_effect = ["none"]

    @staticmethod
    def mutate(input: str) -> str:
        if len(input) == 0:
            return input
        position = Mutator.random_int(len(input))
        replacement_character = chr(ord(input[position]) ^ (1 << Mutator.random_int(8)))
        mutation = input[0:position] + replacement_character + input[position + 1 :]
        return mutation


@register_mutator
class MutatorStringAddSubByte(Mutator):
    allowed_types = set(["str"])
    size_effect = ["none"]

    @staticmethod
    def mutate(input: str) -> str:
        if len(input) == 0:
            return input
        position = Mutator.random_int(len(input))
        delta = Mutator.random_int(256)
        replacement_character = chr((ord(input[position]) + delta) % 256)
        mutation = input[0:position] + replacement_character + input[position + 1 :]
        return mutation


@register_mutator
class MutatorStringInsertCharacter(Mutator):
    allowed_types = set(["str"])
    size_effect = ["increase"]

    @staticmethod
    def mutate(input: str) -> str:
        if len(input) == 0:
            mutation = chr(Mutator.random_int(256))
        position = Mutator.random_int(len(input))
        character = chr(Mutator.random_int(256))
        mutation = input[0:position] + character + input[position:]
        return mutation


@register_mutator
class MutatorStringInsertDict(Mutator):
    allowed_types = set(["str"])
    size_effect = ["increase"]
    # Disable if the dictionary is empty or if the dictionary fuzzing is disabled
    enabled = True if (dictionary and Config.FRELATAGE_DICTIONARY_ENABLE) else False

    @staticmethod
    def mutate(input: str) -> str:
        if len(input) == 0:
            return input
        position = Mutator.random_int(len(input))
        character = random.choice(dictionary)
        mutation = input[0:position] + character + input[position:]
        return mutation


@register_mutator
class MutatorStringDeleteCharacter(Mutator):
    allowed_types = set(["str"])
    size_effect = ["decrease"]

    @staticmethod
    def mutate(input: str) -> str:
        if len(input) == 0:
            return input
        position = Mutator.random_int(len(input))
        mutation = input[: position - 1] + input[position]
        return mutation


@register_mutator
class MutatorStringSwapTwoChars(Mutator):
    allowed_types = set(["str"])
    size_effect = ["none"]

    @staticmethod
    def mutate(input: str) -> str:
        if len(input) < 2:
            return input
        first_character_position = Mutator.random_int(len(input))
        second_character_position = Mutator.random_int(len(input))
        # The position of the second character must be different from the first
        while first_character_position == second_character_position:
            second_character_position = Mutator.random_int(len(input))

        mutation = list(input)
        (mutation[first_character_position], mutation[second_character_position],) = (
            mutation[second_character_position],
            mutation[first_character_position],
        )
        mutation_str = "".join(mutation)
        return mutation_str


@register_mutator
class MutatorStringDuplicateSubString(Mutator):
    allowed_types = set(["str"])
    size_effect = ["increase"]

    @staticmethod
    def mutate(input: str) -> str:
        if len(input) < 2:
            return input
        first_character_position = Mutator.random_int(len(input))
        second_character_position = random.randint(first_character_position, len(input))
        substring = input[first_character_position:second_character_position]

        substring_position = Mutator.random_int(len(input))
        mutation = input[:substring_position] + substring + input[substring_position:]
        return mutation


@register_mutator
class MutatorStringRepeatSubString(Mutator):
    allowed_types = set(["str"])
    size_effect = ["increase"]

    @staticmethod
    def mutate(input: str) -> str:
        if len(input) < 2:
            return input
        first_character_position = Mutator.random_int(len(input))
        second_character_position = random.randint(first_character_position, len(input))
        substring = input[first_character_position:second_character_position]
        mutation = (
            input[:second_character_position]
            + substring
            + input[second_character_position:]
        )
        return mutation


@register_mutator
class MutatorStringDeleteSubString(Mutator):
    allowed_types = set(["str"])
    size_effect = ["decrease"]

    @staticmethod
    def mutate(input: str) -> str:
        if len(input) < 2:
            return input
        first_character_position = Mutator.random_int(len(input))
        second_character_position = random.randint(first_character_position, len(input))

        mutation = input[:first_character_position] + input[second_character_position:]
        return mutation


@register_mutator
class MutatorIntMagicValue(Mutator):
    allowed_types = set(["int"])
    size_effect = ["none", "increase", "decrease"]

    @staticmethod
    def mutate(input: int) -> int:
        mutation = random.choice(MagicValues.UINT)
        return mutation


@register_mutator
class MutatorIntAddSub(Mutator):
    allowed_types = set(["int"])
    size_effect = ["none", "increase", "decrease"]

    @staticmethod
    def mutate(input: int) -> int:
        mutation = input
        mutation += random.randint(-1000, 1000)
        return mutation


# List mutators
@register_mutator
class MutatorListDuplicateElement(Mutator):
    allowed_types = set(["list"])
    size_effect = ["increase"]

    @staticmethod
    def mutate(input: list) -> list:
        if len(input) == 0:
            return input

        mutation = input.copy()
        position = Mutator.random_int(len(mutation))
        element = random.choice(mutation)
        mutation.insert(position, element)
        return mutation


@register_mutator
class MutatorListInsertNone(Mutator):
    allowed_types = set(["list"])
    size_effect = ["increase"]

    @staticmethod
    def mutate(input: list) -> list:
        mutation = input.copy()
        mutation.append(None)
        return mutation


@register_mutator
class MutatorListRemoveElement(Mutator):
    allowed_types = set(["list"])
    size_effect = ["decrease"]

    @staticmethod
    def mutate(input: list) -> list:
        if len(input) == 0:
            return input

        mutation = input.copy()
        element_position = Mutator.random_int(len(mutation))
        mutation.pop(element_position)
        return mutation


@register_mutator
class MutatorListShuffle(Mutator):
    allowed_types = set(["list"])
    size_effect = ["none"]

    @staticmethod
    def mutate(input: list) -> list:
        if len(input) == 0:
            return input

        mutation = input.copy()
        random.shuffle(mutation)
        return mutation


# Tuple mutators
@register_mutator
class MutatorTupleDuplicateElement(Mutator):
    allowed_types = set(["tuple"])
    size_effect = ["increase"]

    @staticmethod
    def mutate(input: tuple) -> tuple:
        if len(input) == 0:
            return input

        mutation = list(input)
        position = Mutator.random_int(len(mutation))
        element = random.choice(mutation)
        mutation.insert(position, element)
        mutation_tuple = tuple(mutation)
        return mutation_tuple


@register_mutator
class MutatorTupleInsertNone(Mutator):
    allowed_types = set(["tuple"])
    size_effect = ["increase"]

    @staticmethod
    def mutate(input: tuple) -> tuple:
        mutation = list(input)
        mutation.append(None)
        mutation_tuple = tuple(mutation)
        return mutation_tuple


@register_mutator
class MutatorTupleRemoveElement(Mutator):
    allowed_types = set(["tuple"])
    size_effect = ["decrease"]

    @staticmethod
    def mutate(input: tuple) -> tuple:
        if len(input) == 0:
            return input

        mutation = list(input)
        element_position = Mutator.random_int(len(mutation))
        mutation.pop(element_position)
        mutation_tuple = tuple(mutation)
        return mutation_tuple


@register_mutator
class MutatorTupleShuffle(Mutator):
    allowed_types = set(["tuple"])
    size_effect = ["none"]

    @staticmethod
    def mutate(input: tuple) -> tuple:
        if len(input) == 0:
            return input

        mutation = list(input)
        random.shuffle(mutation)
        mutation_tuple = tuple(mutation)
        return mutation_tuple


# None mutators
@register_mutator
class MutatorNone(Mutator):
    allowed_types = set(["NoneType"])
    size_effect = ["none", "increase", "decrease"]

    @staticmethod
    def mutate(input: None) -> Any:
        types = ["a", 0, [], {}, 1.0, ()]
        mutation = random.choice(types)
        return mutation


# File mutators
@register_mutator
class MutatorFileFlipBit(Mutator):
    allowed_types = set(["file"])
    size_effect = ["none"]

    @staticmethod
    def mutate(input: str) -> str:
        with open(input, "rb") as f:
            file_content = f.read()
            if len(file_content) == 0:
                return input
            else:
                position = Mutator.random_int(len(file_content))

                mutation = bytearray(file_content)
                mutation[position] = mutation[position] ^ (1 << Mutator.random_int(8))
                mutation_bytes = bytes(mutation)
        with open(input, "wb") as f:
            f.write(mutation_bytes)
        return input


@register_mutator
class MutatorFileInsertByte(Mutator):
    allowed_types = set(["file"])
    size_effect = ["increase"]

    @staticmethod
    def mutate(input: str) -> str:
        with open(input, "rb") as f:
            file_content = f.read()

            mutation: bytes = bytearray(file_content)
            if len(file_content) == 0:
                mutation = chr(Mutator.random_int(256)).encode()
            else:
                position = Mutator.random_int(len(file_content))
                mutation = (
                    mutation[0:position]
                    + chr(Mutator.random_int(256)).encode()
                    + mutation[position:]
                )
            mutation = bytes(mutation)
        with open(input, "wb") as f:
            f.write(mutation)
        return input


@register_mutator
class MutatorFileDeleteByte(Mutator):
    allowed_types = set(["file"])
    size_effect = ["decrease"]

    @staticmethod
    def mutate(input: str) -> str:
        with open(input, "rb") as f:
            file_content = f.read()
            if len(file_content) <= 2:
                return input
            else:
                position = Mutator.random_int(len(file_content))

                mutation: bytes = bytearray(file_content)
                mutation = mutation[0:position] + mutation[position + 1 :]
                mutation = bytes(mutation)
        with open(input, "wb") as f:
            f.write(mutation)
        return input


@register_mutator
class MutatorFileSwapTwoBytes(Mutator):
    allowed_types = set(["file"])
    size_effect = ["none"]

    @staticmethod
    def mutate(input: str) -> str:
        with open(input, "rb") as f:
            file_content = f.read()
            if len(file_content) <= 2:
                return input
            else:
                first_byte_position = Mutator.random_int(len(file_content))
                second_byte_position = Mutator.random_int(len(file_content))
                # The position of the second character must be different from the first
                while first_byte_position == second_byte_position:
                    second_byte_position = Mutator.random_int(len(file_content))

                mutation = bytearray(file_content)
                (mutation[first_byte_position], mutation[second_byte_position],) = (
                    mutation[second_byte_position],
                    mutation[first_byte_position],
                )
                mutation_bytes = bytes(mutation)
        with open(input, "wb") as f:
            f.write(mutation_bytes)
        return input


@register_mutator
class MutatorFileDuplicateSubBytes(Mutator):
    allowed_types = set(["file"])
    size_effect = ["increase"]

    @staticmethod
    def mutate(input: str) -> str:
        with open(input, "rb") as f:
            file_content = f.read()
            if len(file_content) == 0:
                return input
            else:
                first_byte_position = Mutator.random_int(len(file_content))
                second_byte_position = random.randint(
                    first_byte_position, len(file_content)
                )
                mutation = bytearray(file_content)

                subbytes = mutation[first_byte_position:second_byte_position]
                subbytes_position = Mutator.random_int(len(file_content))
                mutation = (
                    mutation[:subbytes_position] + subbytes + mutation[subbytes_position:]
                )
                mutation_bytes = bytes(mutation)
        with open(input, "wb") as f:
            f.write(mutation_bytes)
        return input


@register_mutator
class MutatorFileRepeatSubBytes(Mutator):
    allowed_types = set(["file"])
    size_effect = ["increase"]

    @staticmethod
    def mutate(input: str) -> str:
        with open(input, "rb") as f:
            file_content = f.read()
            if len(file_content) == 0:
                return input
            else:
                first_byte_position = Mutator.random_int(len(file_content))
                second_byte_position = random.randint(
                    first_byte_position, len(file_content)
                )
                mutation = bytearray(file_content)

                subbytes = mutation[first_byte_position:second_byte_position]
                mutation = (
                    mutation[:second_byte_position]
                    + subbytes
                    + mutation[second_byte_position:]
                )
                mutation_bytes = bytes(mutation)
        with open(input, "wb") as f:
            f.write(mutation_bytes)
        return input


@register_mutator
class MutatorFileDeleteSubBytes(Mutator):
    allowed_types = set(["file"])
    size_effect = ["decrease"]

    @staticmethod
    def mutate(input: str) -> str:
        with open(input, "rb") as f:
            file_content = f.read()
            if len(file_content) <= 2:
                return input
            else:
                first_byte_position = Mutator.random_int(len(file_content))
                second_byte_position = random.randint(
                    first_byte_position, len(file_content)
                )
                mutation = bytearray(file_content)
                mutation = (
                    mutation[:first_byte_position] + mutation[second_byte_position:]
                )
                mutation_bytes = bytes(mutation)
        with open(input, "wb") as f:
            f.write(mutation_bytes)
        return input


@register_mutator
class MutatorFileInsertDict(Mutator):
    allowed_types = set(["file"])
    size_effect = ["increase"]
    # Disable if the dictionary is empty or if the dictionary fuzzing is disabled
    enabled = True if (dictionary and Config.FRELATAGE_DICTIONARY_ENABLE) else False

    @staticmethod
    def mutate(input: str) -> str:
        with open(input, "rb") as f:
            file_content = f.read()

            element = random.choice(dictionary)

            mutation = bytearray(file_content)
            if len(file_content) == 0:
                mutation = element.encode()
            else:
                position = Mutator.random_int(len(file_content))
                mutation = mutation[0:position] + element.encode() + mutation[position:]
            mutation_bytes = bytes(mutation)
        with open(input, "wb") as f:
            f.write(mutation_bytes)
        return input


# Dictionnary mutators
@register_mutator
class MutatorDictAddEntry(Mutator):
    allowed_types = set(["dict"])
    size_effect = ["increase"]

    @staticmethod
    def mutate(input: dict) -> dict:
        mutation = input

        if random.randint(0, 1) == 0:
            integer = random.choice(MagicValues.UINT)
            mutation[integer] = None
        else:
            string = chr(Mutator.random_int(256))
            mutation[string] = None
        return mutation


@register_mutator
class MutatorDictDeleteEntry(Mutator):
    allowed_types = set(["dict"])
    size_effect = ["decrease"]

    @staticmethod
    def mutate(input: dict) -> dict:
        if len(input) == 0:
            return input
        mutation = input
        del mutation[random.choice(list(mutation.keys()))]
        return mutation


# Float mutators
@register_mutator
class MutatorFloatAddSub(Mutator):
    allowed_types = set(["float"])
    size_effect = ["none", "increase", "decrease"]

    @staticmethod
    def mutate(input: float) -> float:
        digits = Mutator.random_int(10)
        delta = random.uniform(-1000, 1000)
        mutation = round(input + delta, digits)
        return mutation



import copy
import numpy as np
from frelatage.input.input import Input


class Queue:
    """
    The queue contains the list of possible combinations of the corpus entries.
    """

    def __init__(self, corpus: list[object]):
        self.position = 0
        self.corpus = corpus
        self.arguments = self.generate_arguments()
        self.currents_arguments = self.current_arguments()
        self.size = len(self.arguments)

    def generate_arguments(self):
        """
        Return all possible combinations of inputs.
        Corpus : [  [Input_1, Input_2],  [Input_3, Input_4]  ]
        Arguments : [
                      [Input_1, Input_3],
                      [Input_1, Input_4],
                      [Input_2, Input_3],
                      [Input_2, Input_4],
                    ]
        """
        arguments_list = np.array(np.meshgrid(*self.corpus)).T.reshape(
            -1, len(self.corpus)
        )
        # Deepcopy each Input instance to avoid collisions
        for i in range(len(arguments_list)):
            arguments = arguments_list[i]
            for j in range(len(arguments)):
                arguments_list[i][j] = copy.deepcopy(arguments_list[i][j])

        arguments = [list(arguments) for arguments in arguments_list]
        return arguments

    def current_arguments(self) -> list[Input]:
        """
        Return the current arguments
        """
        return self.arguments[self.position]

    @property
    def end(self) -> bool:
        """
        Check if we have reached the end of the queue
        """
        return self.position >= len(self.arguments)



from typing import Any


class Report:
    def __init__(
        self,
        trace: Any,
        input: list,
        error: bool,
        timeout: bool,
        new_error_instruction: bool,
        reached_instructions_count: int,
        new_reached_instructions_count: int,
        instructions_pairs_count: int,
        new_instructions_pairs_count: int,
    ) -> None:
        """
        Report of the behavior of a function with given input
        """
        self.trace = trace
        self.input = input
        self.error = error
        self.timeout = timeout
        self.new_error_instruction = new_error_instruction
        self.reached_instructions_count = reached_instructions_count
        self.new_reached_instructions_count = new_reached_instructions_count
        self.instructions_pairs_count = instructions_pairs_count
        self.new_instructions_pairs_count = new_instructions_pairs_count

        self.score = self.compute_score()

    def compute_score(self) -> int:
        """
        The score of a report is used to select the most efficient inputs for the fuzzer's genetic algorithm mutations.
        The parameters are : the number of instructions reached, the number of new instructions reached, the number of
        pairs of instructions reached, the occurrence of an error/timeout.
        """

        result = (
            self.reached_instructions_count
            + self.new_reached_instructions_count
            + self.instructions_pairs_count
            + self.new_instructions_pairs_count
        )
        return result

    def __eq__(self, other):
        """
        Reports are sorted by score
        """
        return self.score == other.score

    def __lt__(self, other):
        """
        Reports are sorted by score
        """
        return self.score < other.score



import contextlib
import os
import timeout_decorator
import trace
from typing import Callable
from frelatage.config.config import Config
from frelatage.tracer.trace_result import Result


def supress_stdout(func: Callable) -> Callable:
    """
    redirect the output of a function to /dev/null
    """

    def wrapper(*a, **ka):
        with open(os.devnull, "w") as devnull:
            with contextlib.redirect_stdout(devnull):
                return func(*a, **ka)

    return wrapper


class Tracer:
    """
    The tracer allows to launch a function and to keep track of the executed instructions
    as well as the triggered errors.
    """

    def __init__(
        self,
        exceptions_whitelist: tuple = (),
        exceptions_blacklist: tuple = (),
    ) -> None:
        """
        Initialize the tracer with an exception whitelist and an exception blacklist.
        """
        self.exceptions_whitelist = exceptions_whitelist
        self.exceptions_blacklist = exceptions_blacklist

    @timeout_decorator.timeout(
        Config.FRELATAGE_TIMEOUT_DELAY,
        use_signals=True,
        timeout_exception=TimeoutError,
    )
    def runfunc(self, function: Callable, arguments):
        """
        Trace a function.
        The result will be stored in self.tracer.
        """
        self.tracer.runfunc(function, *arguments)

    @supress_stdout
    def trace(self, function: Callable, arguments: list) -> Result:
        """
        Trace a function and return a trace result object.
        """
        error = False
        error_type = None
        timeout = False

        # Initialize the tracer
        self.tracer = trace.Trace(ignoredirs=[], trace=0, count=1)

        arguments = [argument.value for argument in arguments]

        # Run the function and catch errors
        try:
            self.runfunc(function, arguments)
        # Whitelisted exceptions
        except self.exceptions_whitelist as e:
            error = True
            error_type = str(e.__class__.__name__)
        # Timeout exception
        except TimeoutError as e:
            timeout = True
            error_type = str(e.__class__.__name__)
        # Blacklisted exceptions
        except self.exceptions_blacklist:
            pass
        # Default exceptions
        except Exception as e:
            if self.exceptions_whitelist == ():
                error = True
                error_type = str(e.__class__.__name__)
            else:
                pass

        # List of the executed instructions
        instructions = list(self.tracer.results().counter.items())  # type: ignore
        reached_instructions = [
            (instruction[0][0], instruction[0][1]) for instruction in instructions
        ]
        # List of the instructions pairs :
        # e.g : reached instructions : [('library.py', 1), ('library.py', 3), ('library.py', 2), ('library.py', 4)]
        #       instructions pairs -> [(('library.py', 3), ('library.py', 3)), (('library.py', 3), ('library.py', 2)), ...]
        instructions_pairs = []
        for i in range(len(reached_instructions)):
            instructions_pairs.append(
                (reached_instructions[i - 1], reached_instructions[i])
            )
        # Position of the instruction where the error occured, None if no error
        if error or timeout:
            error_position = (reached_instructions[-1], error_type)
        else:
            error_position = None

        # Generate a result object
        result = Result(
            error,
            error_type,
            timeout,
            error_position,
            reached_instructions,
            instructions_pairs,
        )
        return result



import math
import os
import sys


class FrelatageConfigError(Exception):
    pass


class Config:
    """
    We can pass configuration through environnement variables
    """

    # Enable the use of mutations based on dictionary elements
    if os.getenv("FRELATAGE_DICTIONARY_ENABLE", "1") not in ("1", "0"):
        raise FrelatageConfigError("FRELATAGE_DICTIONARY_ENABLE must '1' or '0'")
    FRELATAGE_DICTIONARY_ENABLE = (
        True if os.getenv("FRELATAGE_DICTIONARY_ENABLE", "1") == "1" else False
    )

    # Save coverage increase reports to reuse them later
    if os.getenv("FRELATAGE_SAVE_NEW_COVERAGE", "1") not in ("1", "0"):
        raise FrelatageConfigError("FRELATAGE_SAVE_NEW_COVERAGE must '1' or '0'")
    FRELATAGE_SAVE_NEW_COVERAGE = (
        True if os.getenv("FRELATAGE_SAVE_NEW_COVERAGE", "1") == "1" else False
    )

    # Enable the debug mode
    if os.getenv("FRELATAGE_DEBUG_MODE", "1") not in ("1", "0"):
        raise FrelatageConfigError("FRELATAGE_DEBUG_MODE must '1' or '0'")
    FRELATAGE_DEBUG_MODE = (
        True if os.getenv("FRELATAGE_DEBUG_MODE", "1") == "1" else False
    )

    # Delay in seconds after which a function will return a timeoutError
    # Must be greater than 1
    if not 1 <= int(os.getenv("FRELATAGE_TIMEOUT_DELAY", 2)) <= math.inf:
        raise FrelatageConfigError("FRELATAGE_TIMEOUT_DELAY must be greater than 1")
    FRELATAGE_TIMEOUT_DELAY = int(os.getenv("FRELATAGE_TIMEOUT_DELAY", 2))

    # Temporary folder where input files are stored.
    # Need to be an absolute path
    FRELATAGE_INPUT_FILE_TMP_DIR = str(
        os.getenv("FRELATAGE_INPUT_FILE_TMP_DIR", "/tmp/frelatage")
    )

    # Maximum size of an input variable in bytes
    # Must be greater than 4
    if not 4 <= int(os.getenv("FRELATAGE_INPUT_MAX_LEN", 1024)) <= math.inf:
        raise FrelatageConfigError("FRELATAGE_INPUT_MAX_LEN must be greater than 4")
    FRELATAGE_INPUT_MAX_LEN = int(os.getenv("FRELATAGE_INPUT_MAX_LEN", 1024))

    # Maximum number of simultaneous threads
    # Must be greater than 8
    if not 8 <= int(os.getenv("FRELATAGE_MAX_THREADS", 20)) <= math.inf:
        raise FrelatageConfigError("FRELATAGE_MAX_THREADS must be greater than 8")
    FRELATAGE_MAX_THREADS = int(os.getenv("FRELATAGE_MAX_THREADS", 20))

    # Maximum number of stages for a fuzzed function
    # Must be in range greater than 1
    if not 1 <= int(os.getenv("FRELATAGE_MAX_STAGES", 1000000)) <= math.inf:
        raise FrelatageConfigError("FRELATAGE_MAX_STAGES must be greater than 1")
    FRELATAGE_MAX_STAGES = int(os.getenv("FRELATAGE_MAX_STAGES", 1000000))

    # Maximum number of successives cycle without a new path
    # Must be greater than 10
    if (
        not 10
        <= int(os.getenv("FRELATAGE_MAX_CYCLES_WITHOUT_NEW_PATHS", 5000))
        <= math.inf
    ):
        raise FrelatageConfigError(
            "FRELATAGE_MAX_CYCLES_WITHOUT_NEW_PATHS must be in range 10~10000"
        )
    FRELATAGE_MAX_CYCLES_WITHOUT_NEW_PATHS = int(
        os.getenv("FRELATAGE_MAX_CYCLES_WITHOUT_NEW_PATHS", 5000)
    )

    # Default directory for dictionaries
    # Relative path (to the path of the fuzzing file)
    FRELATAGE_DICTIONARY_DIR = str(
        os.path.join(
            os.path.dirname(os.path.realpath(sys.argv[0])),
            os.getenv("FRELATAGE_DICTIONARY_DIR", "./dict"),
        )
    )

    # Default directory for input files
    # Relative path (to the path of the fuzzing file)
    FRELATAGE_INPUT_DIR = str(
        os.path.join(
            os.path.dirname(os.path.realpath(sys.argv[0])),
            os.getenv("FRELATAGE_INPUT_DIR", "./in"),
        )
    )
